@@ -1,4 +1,4 @@
-# u-msgpack-python v1.8 - vsergeev at gmail
+# u-msgpack-python v2.0 - vsergeev at gmail
 # https://github.com/vsergeev/u-msgpack-python
 #
 # u-msgpack-python is a lightweight MessagePack serializer and deserializer
@@ -31,7 +31,7 @@
 # THE SOFTWARE.
 #
 """
-u-msgpack-python v1.8 - vsergeev at gmail
+u-msgpack-python v2.0 - vsergeev at gmail
 https://github.com/vsergeev/u-msgpack-python
 
 u-msgpack-python is a lightweight MessagePack serializer and deserializer
@@ -44,20 +44,23 @@ types.
 License: MIT
 """
 
-version = (1,8)
+version = (2,0)
 "Module version tuple"
 
 import struct
 import collections
 import sys
+import io
 
+################################################################################
+### Ext Class
 ################################################################################
 
-# Extension type for application-specific types and data
+# Extension type for application-defined types and data
 class Ext:
     """
     The Ext class facilitates creating a serializable extension object to store
-    an application-specific type and data byte array.
+    an application-defined type and data byte array.
     """
 
     def __init__(self, type, data):
@@ -65,8 +68,8 @@ class Ext:
         Construct a new Ext object.
 
         Args:
-            type: application-specific type integer from 0 to 127
-            data: application-specific data byte array
+            type: application-defined type integer from 0 to 127
+            data: application-defined data byte array
 
         Raises:
             TypeError:
@@ -124,6 +127,8 @@ class Ext:
         return s
 
 ################################################################################
+### Exceptions
+################################################################################
 
 # Base Exception classes
 class PackException(Exception):
@@ -162,12 +167,18 @@ class DuplicateKeyException(UnpackException):
 KeyNotPrimitiveException = UnhashableKeyException
 KeyDuplicateException = DuplicateKeyException
 
+################################################################################
+### Exported Functions and Globals
 ################################################################################
 
-# Exported functions and variables set in __init()
+# Exported functions and variables, set up in __init()
+pack = None
 packb = None
+unpack = None
 unpackb = None
+dump = None
 dumps = None
+load = None
 loads = None
 
 compatibility = False
@@ -189,147 +200,243 @@ b'\x92\xabsome string\xaasome bytes'
 >>>
 """
 
+################################################################################
+### Packing
 ################################################################################
 
-# You may notice struct.pack("B", x) instead of the simpler chr(x) in the code
-# below. This is to allow for seamless Python 2 and 3 compatibility, as chr(x)
-# has a str return type instead of bytes in Python 3, and struct.pack(...) has
-# the right return type in both versions.
+# You may notice struct.pack("B", obj) instead of the simpler chr(obj) in the
+# code below. This is to allow for seamless Python 2 and 3 compatibility, as
+# chr(obj) has a str return type instead of bytes in Python 3, and
+# struct.pack(...) has the right return type in both versions.
 
-def _pack_integer(x):
-    if x < 0:
-        if x >= -32:
-            return struct.pack("b", x)
-        elif x >= -2**(8-1):
-            return b"\xd0" + struct.pack("b", x)
-        elif x >= -2**(16-1):
-            return b"\xd1" + struct.pack(">h", x)
-        elif x >= -2**(32-1):
-            return b"\xd2" + struct.pack(">i", x)
-        elif x >= -2**(64-1):
-            return b"\xd3" + struct.pack(">q", x)
+def _pack_integer(obj, fp):
+    if obj < 0:
+        if obj >= -32:
+            fp.write(struct.pack("b", obj))
+        elif obj >= -2**(8-1):
+            fp.write(b"\xd0" + struct.pack("b", obj))
+        elif obj >= -2**(16-1):
+            fp.write(b"\xd1" + struct.pack(">h", obj))
+        elif obj >= -2**(32-1):
+            fp.write(b"\xd2" + struct.pack(">i", obj))
+        elif obj >= -2**(64-1):
+            fp.write(b"\xd3" + struct.pack(">q", obj))
         else:
             raise UnsupportedTypeException("huge signed int")
     else:
-        if x <= 127:
-            return struct.pack("B", x)
-        elif x <= 2**8-1:
-            return b"\xcc" + struct.pack("B", x)
-        elif x <= 2**16-1:
-            return b"\xcd" + struct.pack(">H", x)
-        elif x <= 2**32-1:
-            return b"\xce" + struct.pack(">I", x)
-        elif x <= 2**64-1:
-            return b"\xcf" + struct.pack(">Q", x)
+        if obj <= 127:
+            fp.write(struct.pack("B", obj))
+        elif obj <= 2**8-1:
+            fp.write(b"\xcc" + struct.pack("B", obj))
+        elif obj <= 2**16-1:
+            fp.write(b"\xcd" + struct.pack(">H", obj))
+        elif obj <= 2**32-1:
+            fp.write(b"\xce" + struct.pack(">I", obj))
+        elif obj <= 2**64-1:
+            fp.write(b"\xcf" + struct.pack(">Q", obj))
         else:
             raise UnsupportedTypeException("huge unsigned int")
 
-def _pack_nil(x):
-    return b"\xc0"
+def _pack_nil(obj, fp):
+    fp.write(b"\xc0")
 
-def _pack_boolean(x):
-    return b"\xc3" if x else b"\xc2"
+def _pack_boolean(obj, fp):
+    fp.write(b"\xc3" if obj else b"\xc2")
 
-def _pack_float(x):
+def _pack_float(obj, fp):
     if _float_size == 64:
-        return b"\xcb" + struct.pack(">d", x)
+        fp.write(b"\xcb" + struct.pack(">d", obj))
     else:
-        return b"\xca" + struct.pack(">f", x)
+        fp.write(b"\xca" + struct.pack(">f", obj))
 
-def _pack_string(x):
-    x = x.encode('utf-8')
-    if len(x) <= 31:
-        return struct.pack("B", 0xa0 | len(x)) + x
-    elif len(x) <= 2**8-1:
-        return b"\xd9" + struct.pack("B", len(x)) + x
-    elif len(x) <= 2**16-1:
-        return b"\xda" + struct.pack(">H", len(x)) + x
-    elif len(x) <= 2**32-1:
-        return b"\xdb" + struct.pack(">I", len(x)) + x
+def _pack_string(obj, fp):
+    obj = obj.encode('utf-8')
+    if len(obj) <= 31:
+        fp.write(struct.pack("B", 0xa0 | len(obj)) + obj)
+    elif len(obj) <= 2**8-1:
+        fp.write(b"\xd9" + struct.pack("B", len(obj)) + obj)
+    elif len(obj) <= 2**16-1:
+        fp.write(b"\xda" + struct.pack(">H", len(obj)) + obj)
+    elif len(obj) <= 2**32-1:
+        fp.write(b"\xdb" + struct.pack(">I", len(obj)) + obj)
     else:
         raise UnsupportedTypeException("huge string")
 
-def _pack_binary(x):
-    if len(x) <= 2**8-1:
-        return b"\xc4" + struct.pack("B", len(x)) + x
-    elif len(x) <= 2**16-1:
-        return b"\xc5" + struct.pack(">H", len(x)) + x
-    elif len(x) <= 2**32-1:
-        return b"\xc6" + struct.pack(">I", len(x)) + x
+def _pack_binary(obj, fp):
+    if len(obj) <= 2**8-1:
+        fp.write(b"\xc4" + struct.pack("B", len(obj)) + obj)
+    elif len(obj) <= 2**16-1:
+        fp.write(b"\xc5" + struct.pack(">H", len(obj)) + obj)
+    elif len(obj) <= 2**32-1:
+        fp.write(b"\xc6" + struct.pack(">I", len(obj)) + obj)
     else:
         raise UnsupportedTypeException("huge binary string")
 
-def _pack_oldspec_raw(x):
-    if len(x) <= 31:
-        return struct.pack("B", 0xa0 | len(x)) + x
-    elif len(x) <= 2**16-1:
-        return b"\xda" + struct.pack(">H", len(x)) + x
-    elif len(x) <= 2**32-1:
-        return b"\xdb" + struct.pack(">I", len(x)) + x
+def _pack_oldspec_raw(obj, fp):
+    if len(obj) <= 31:
+        fp.write(struct.pack("B", 0xa0 | len(obj)) + obj)
+    elif len(obj) <= 2**16-1:
+        fp.write(b"\xda" + struct.pack(">H", len(obj)) + obj)
+    elif len(obj) <= 2**32-1:
+        fp.write(b"\xdb" + struct.pack(">I", len(obj)) + obj)
     else:
         raise UnsupportedTypeException("huge raw string")
 
-def _pack_ext(x):
-    if len(x.data) == 1:
-        return b"\xd4" + struct.pack("B", x.type & 0xff) + x.data
-    elif len(x.data) == 2:
-        return b"\xd5" + struct.pack("B", x.type & 0xff) + x.data
-    elif len(x.data) == 4:
-        return b"\xd6" + struct.pack("B", x.type & 0xff) + x.data
-    elif len(x.data) == 8:
-        return b"\xd7" + struct.pack("B", x.type & 0xff) + x.data
-    elif len(x.data) == 16:
-        return b"\xd8" + struct.pack("B", x.type & 0xff) + x.data
-    elif len(x.data) <= 2**8-1:
-        return b"\xc7" + struct.pack("BB", len(x.data), x.type & 0xff) + x.data
-    elif len(x.data) <= 2**16-1:
-        return b"\xc8" + struct.pack(">HB", len(x.data), x.type & 0xff) + x.data
-    elif len(x.data) <= 2**32-1:
-        return b"\xc9" + struct.pack(">IB", len(x.data), x.type & 0xff) + x.data
+def _pack_ext(obj, fp):
+    if len(obj.data) == 1:
+        fp.write(b"\xd4" + struct.pack("B", obj.type & 0xff) + obj.data)
+    elif len(obj.data) == 2:
+        fp.write(b"\xd5" + struct.pack("B", obj.type & 0xff) + obj.data)
+    elif len(obj.data) == 4:
+        fp.write(b"\xd6" + struct.pack("B", obj.type & 0xff) + obj.data)
+    elif len(obj.data) == 8:
+        fp.write(b"\xd7" + struct.pack("B", obj.type & 0xff) + obj.data)
+    elif len(obj.data) == 16:
+        fp.write(b"\xd8" + struct.pack("B", obj.type & 0xff) + obj.data)
+    elif len(obj.data) <= 2**8-1:
+        fp.write(b"\xc7" + struct.pack("BB", len(obj.data), obj.type & 0xff) + obj.data)
+    elif len(obj.data) <= 2**16-1:
+        fp.write(b"\xc8" + struct.pack(">HB", len(obj.data), obj.type & 0xff) + obj.data)
+    elif len(obj.data) <= 2**32-1:
+        fp.write(b"\xc9" + struct.pack(">IB", len(obj.data), obj.type & 0xff) + obj.data)
     else:
         raise UnsupportedTypeException("huge ext data")
 
-def _pack_array(x):
-    if len(x) <= 15:
-        s = struct.pack("B", 0x90 | len(x))
-    elif len(x) <= 2**16-1:
-        s = b"\xdc" + struct.pack(">H", len(x))
-    elif len(x) <= 2**32-1:
-        s = b"\xdd" + struct.pack(">I", len(x))
+def _pack_array(obj, fp):
+    if len(obj) <= 15:
+        fp.write(struct.pack("B", 0x90 | len(obj)))
+    elif len(obj) <= 2**16-1:
+        fp.write(b"\xdc" + struct.pack(">H", len(obj)))
+    elif len(obj) <= 2**32-1:
+        fp.write(b"\xdd" + struct.pack(">I", len(obj)))
     else:
         raise UnsupportedTypeException("huge array")
 
-    for e in x:
-        s += packb(e)
+    for e in obj:
+        pack(e, fp)
 
-    return s
-
-def _pack_map(x):
-    if len(x) <= 15:
-        s = struct.pack("B", 0x80 | len(x))
-    elif len(x) <= 2**16-1:
-        s = b"\xde" + struct.pack(">H", len(x))
-    elif len(x) <= 2**32-1:
-        s = b"\xdf" + struct.pack(">I", len(x))
+def _pack_map(obj, fp):
+    if len(obj) <= 15:
+        fp.write(struct.pack("B", 0x80 | len(obj)))
+    elif len(obj) <= 2**16-1:
+        fp.write(b"\xde" + struct.pack(">H", len(obj)))
+    elif len(obj) <= 2**32-1:
+        fp.write(b"\xdf" + struct.pack(">I", len(obj)))
     else:
         raise UnsupportedTypeException("huge array")
 
-    for k,v in x.items():
-        s += packb(k)
-        s += packb(v)
+    for k,v in obj.items():
+        pack(k, fp)
+        pack(v, fp)
 
-    return s
+########################################
 
 # Pack for Python 2, with 'unicode' type, 'str' type, and 'long' type
-def _packb2(x):
+def _pack2(obj, fp):
     """
     Serialize a Python object into MessagePack bytes.
 
     Args:
-        x: Python object
+        obj: a Python object
+        fp: a .write()-supporting file-like object
 
     Returns:
-        A 'str' containing the serialized bytes.
+        None.
+
+    Raises:
+        UnsupportedType(PackException):
+            Object type not supported for packing.
+
+    Example:
+    >>> f = open('test.bin', 'w')
+    >>> umsgpack.pack({u"compact": True, u"schema": 0}, f)
+    >>>
+    """
+
+    global compatibility
+
+    if obj is None:
+        _pack_nil(obj, fp)
+    elif isinstance(obj, bool):
+        _pack_boolean(obj, fp)
+    elif isinstance(obj, int) or isinstance(obj, long):
+        _pack_integer(obj, fp)
+    elif isinstance(obj, float):
+        _pack_float(obj, fp)
+    elif compatibility and isinstance(obj, unicode):
+        _pack_oldspec_raw(bytes(obj), fp)
+    elif compatibility and isinstance(obj, bytes):
+        _pack_oldspec_raw(obj, fp)
+    elif isinstance(obj, unicode):
+        _pack_string(obj, fp)
+    elif isinstance(obj, str):
+        _pack_binary(obj, fp)
+    elif isinstance(obj, list) or isinstance(obj, tuple):
+        _pack_array(obj, fp)
+    elif isinstance(obj, dict):
+        _pack_map(obj, fp)
+    elif isinstance(obj, Ext):
+        _pack_ext(obj, fp)
+    else:
+        raise UnsupportedTypeException("unsupported type: %s" % str(type(obj)))
+
+# Pack for Python 3, with unicode 'str' type, 'bytes' type, and no 'long' type
+def _pack3(obj, fp):
+    """
+    Serialize a Python object into MessagePack bytes.
+
+    Args:
+        obj: a Python object
+        fp: a .write()-supporting file-like object
+
+    Returns:
+        None.
+
+    Raises:
+        UnsupportedType(PackException):
+            Object type not supported for packing.
+
+    Example:
+    >>> f = open('test.bin', 'w')
+    >>> umsgpack.pack({u"compact": True, u"schema": 0}, fp)
+    >>>
+    """
+    global compatibility
+
+    if obj is None:
+        _pack_nil(obj, fp)
+    elif isinstance(obj, bool):
+        _pack_boolean(obj, fp)
+    elif isinstance(obj, int):
+        _pack_integer(obj, fp)
+    elif isinstance(obj, float):
+        _pack_float(obj, fp)
+    elif compatibility and isinstance(obj, str):
+        _pack_oldspec_raw(obj.encode('utf-8'), fp)
+    elif compatibility and isinstance(obj, bytes):
+        _pack_oldspec_raw(obj, fp)
+    elif isinstance(obj, str):
+        _pack_string(obj, fp)
+    elif isinstance(obj, bytes):
+        _pack_binary(obj, fp)
+    elif isinstance(obj, list) or isinstance(obj, tuple):
+        _pack_array(obj, fp)
+    elif isinstance(obj, dict):
+        _pack_map(obj, fp)
+    elif isinstance(obj, Ext):
+        _pack_ext(obj, fp)
+    else:
+        raise UnsupportedTypeException("unsupported type: %s" % str(type(obj)))
+
+def _packb2(obj):
+    """
+    Serialize a Python object into MessagePack bytes.
+
+    Args:
+        obj: a Python object
+
+    Returns:
+        A 'str' containing serialized MessagePack bytes.
 
     Raises:
         UnsupportedType(PackException):
@@ -340,43 +447,19 @@ def _packb2(x):
     '\x82\xa7compact\xc3\xa6schema\x00'
     >>>
     """
-    global compatibility
+    fp = io.BytesIO()
+    _pack2(obj, fp)
+    return fp.getvalue()
 
-    if x is None:
-        return _pack_nil(x)
-    elif isinstance(x, bool):
-        return _pack_boolean(x)
-    elif isinstance(x, int) or isinstance(x, long):
-        return _pack_integer(x)
-    elif isinstance(x, float):
-        return _pack_float(x)
-    elif compatibility and isinstance(x, unicode):
-        return _pack_oldspec_raw(bytes(x))
-    elif compatibility and isinstance(x, bytes):
-        return _pack_oldspec_raw(x)
-    elif isinstance(x, unicode):
-        return _pack_string(x)
-    elif isinstance(x, str):
-        return _pack_binary(x)
-    elif isinstance(x, list) or isinstance(x, tuple):
-        return _pack_array(x)
-    elif isinstance(x, dict):
-        return _pack_map(x)
-    elif isinstance(x, Ext):
-        return _pack_ext(x)
-    else:
-        raise UnsupportedTypeException("unsupported type: %s" % str(type(x)))
-
-# Pack for Python 3, with unicode 'str' type, 'bytes' type, and no 'long' type
-def _packb3(x):
+def _packb3(obj):
     """
     Serialize a Python object into MessagePack bytes.
 
     Args:
-        x: Python object
+        obj: a Python object
 
     Returns:
-        A 'bytes' containing the serialized bytes.
+        A 'bytes' containing serialized MessagePack bytes.
 
     Raises:
         UnsupportedType(PackException):
@@ -387,117 +470,102 @@ def _packb3(x):
     b'\x82\xa7compact\xc3\xa6schema\x00'
     >>>
     """
-    global compatibility
-
-    if x is None:
-        return _pack_nil(x)
-    elif isinstance(x, bool):
-        return _pack_boolean(x)
-    elif isinstance(x, int):
-        return _pack_integer(x)
-    elif isinstance(x, float):
-        return _pack_float(x)
-    elif compatibility and isinstance(x, str):
-        return _pack_oldspec_raw(x.encode('utf-8'))
-    elif compatibility and isinstance(x, bytes):
-        return _pack_oldspec_raw(x)
-    elif isinstance(x, str):
-        return _pack_string(x)
-    elif isinstance(x, bytes):
-        return _pack_binary(x)
-    elif isinstance(x, list) or isinstance(x, tuple):
-        return _pack_array(x)
-    elif isinstance(x, dict):
-        return _pack_map(x)
-    elif isinstance(x, Ext):
-        return _pack_ext(x)
-    else:
-        raise UnsupportedTypeException("unsupported type: %s" % str(type(x)))
+    fp = io.BytesIO()
+    _pack3(obj, fp)
+    return fp.getvalue()
 
 ################################################################################
+### Unpacking
+################################################################################
 
-def _unpack_integer(code, read_fn):
+def _read_except(fp, n):
+    data = fp.read(n)
+    if len(data) < n:
+        raise InsufficientDataException()
+    return data
+
+def _unpack_integer(code, fp):
     if (ord(code) & 0xe0) == 0xe0:
         return struct.unpack("b", code)[0]
     elif code == b'\xd0':
-        return struct.unpack("b", read_fn(1))[0]
+        return struct.unpack("b", _read_except(fp, 1))[0]
     elif code == b'\xd1':
-        return struct.unpack(">h", read_fn(2))[0]
+        return struct.unpack(">h", _read_except(fp, 2))[0]
     elif code == b'\xd2':
-        return struct.unpack(">i", read_fn(4))[0]
+        return struct.unpack(">i", _read_except(fp, 4))[0]
     elif code == b'\xd3':
-        return struct.unpack(">q", read_fn(8))[0]
+        return struct.unpack(">q", _read_except(fp, 8))[0]
     elif (ord(code) & 0x80) == 0x00:
         return struct.unpack("B", code)[0]
     elif code == b'\xcc':
-        return struct.unpack("B", read_fn(1))[0]
+        return struct.unpack("B", _read_except(fp, 1))[0]
     elif code == b'\xcd':
-        return struct.unpack(">H", read_fn(2))[0]
+        return struct.unpack(">H", _read_except(fp, 2))[0]
     elif code == b'\xce':
-        return struct.unpack(">I", read_fn(4))[0]
+        return struct.unpack(">I", _read_except(fp, 4))[0]
     elif code == b'\xcf':
-        return struct.unpack(">Q", read_fn(8))[0]
+        return struct.unpack(">Q", _read_except(fp, 8))[0]
     raise Exception("logic error, not int: 0x%02x" % ord(code))
 
-def _unpack_reserved(code, read_fn):
+def _unpack_reserved(code, fp):
     if code == b'\xc1':
         raise ReservedCodeException("encountered reserved code: 0x%02x" % ord(code))
     raise Exception("logic error, not reserved code: 0x%02x" % ord(code))
 
-def _unpack_nil(code, read_fn):
+def _unpack_nil(code, fp):
     if code == b'\xc0':
         return None
     raise Exception("logic error, not nil: 0x%02x" % ord(code))
 
-def _unpack_boolean(code, read_fn):
+def _unpack_boolean(code, fp):
     if code == b'\xc2':
         return False
     elif code == b'\xc3':
         return True
     raise Exception("logic error, not boolean: 0x%02x" % ord(code))
 
-def _unpack_float(code, read_fn):
+def _unpack_float(code, fp):
     if code == b'\xca':
-        return struct.unpack(">f", read_fn(4))[0]
+        return struct.unpack(">f", _read_except(fp, 4))[0]
     elif code == b'\xcb':
-        return struct.unpack(">d", read_fn(8))[0]
+        return struct.unpack(">d", _read_except(fp, 8))[0]
     raise Exception("logic error, not float: 0x%02x" % ord(code))
 
-def _unpack_string(code, read_fn):
+def _unpack_string(code, fp):
     if (ord(code) & 0xe0) == 0xa0:
         length = ord(code) & ~0xe0
     elif code == b'\xd9':
-        length = struct.unpack("B", read_fn(1))[0]
+        length = struct.unpack("B", _read_except(fp, 1))[0]
     elif code == b'\xda':
-        length = struct.unpack(">H", read_fn(2))[0]
+        length = struct.unpack(">H", _read_except(fp, 2))[0]
     elif code == b'\xdb':
-        length = struct.unpack(">I", read_fn(4))[0]
+        length = struct.unpack(">I", _read_except(fp, 4))[0]
     else:
         raise Exception("logic error, not string: 0x%02x" % ord(code))
 
     # Always return raw bytes in compatibility mode
     global compatibility
     if compatibility:
-        return read_fn(length)
+        return _read_except(fp, length)
 
     try:
-        return bytes.decode(read_fn(length), 'utf-8')
+        return bytes.decode(_read_except(fp, length), 'utf-8')
     except UnicodeDecodeError:
         raise InvalidStringException("unpacked string is not utf-8")
 
-def _unpack_binary(code, read_fn):
+def _unpack_binary(code, fp):
     if code == b'\xc4':
-        length = struct.unpack("B", read_fn(1))[0]
+        length = struct.unpack("B", _read_except(fp, 1))[0]
     elif code == b'\xc5':
-        length = struct.unpack(">H", read_fn(2))[0]
+        length = struct.unpack(">H", _read_except(fp, 2))[0]
     elif code == b'\xc6':
-        length = struct.unpack(">I", read_fn(4))[0]
+        length = struct.unpack(">I", _read_except(fp, 4))[0]
     else:
         raise Exception("logic error, not binary: 0x%02x" % ord(code))
 
-    return read_fn(length)
+    return _read_except(fp, length)
 
-def _unpack_ext(code, read_fn):
+def _unpack_ext(code, fp):
     if code == b'\xd4':
         length = 1
     elif code == b'\xd5':
@@ -509,47 +577,47 @@ def _unpack_ext(code, read_fn):
     elif code == b'\xd8':
         length = 16
     elif code == b'\xc7':
-        length = struct.unpack("B", read_fn(1))[0]
+        length = struct.unpack("B", _read_except(fp, 1))[0]
     elif code == b'\xc8':
-        length = struct.unpack(">H", read_fn(2))[0]
+        length = struct.unpack(">H", _read_except(fp, 2))[0]
     elif code == b'\xc9':
-        length = struct.unpack(">I", read_fn(4))[0]
+        length = struct.unpack(">I", _read_except(fp, 4))[0]
     else:
         raise Exception("logic error, not ext: 0x%02x" % ord(code))
 
-    return Ext(ord(read_fn(1)), read_fn(length))
+    return Ext(ord(_read_except(fp, 1)), _read_except(fp, length))
 
-def _unpack_array(code, read_fn):
+def _unpack_array(code, fp):
     if (ord(code) & 0xf0) == 0x90:
         length = (ord(code) & ~0xf0)
     elif code == b'\xdc':
-        length = struct.unpack(">H", read_fn(2))[0]
+        length = struct.unpack(">H", _read_except(fp, 2))[0]
     elif code == b'\xdd':
-        length = struct.unpack(">I", read_fn(4))[0]
+        length = struct.unpack(">I", _read_except(fp, 4))[0]
     else:
         raise Exception("logic error, not array: 0x%02x" % ord(code))
 
-    return [_unpackb(read_fn) for i in range(length)]
+    return [_unpack(fp) for i in range(length)]
 
-def _deep_list_to_tuple(x):
-    if isinstance(x, list):
-        return tuple([_deep_list_to_tuple(e) for e in x])
-    return x
+def _deep_list_to_tuple(obj):
+    if isinstance(obj, list):
+        return tuple([_deep_list_to_tuple(e) for e in obj])
+    return obj
 
-def _unpack_map(code, read_fn):
+def _unpack_map(code, fp):
     if (ord(code) & 0xf0) == 0x80:
         length = (ord(code) & ~0xf0)
     elif code == b'\xde':
-        length = struct.unpack(">H", read_fn(2))[0]
+        length = struct.unpack(">H", _read_except(fp, 2))[0]
     elif code == b'\xdf':
-        length = struct.unpack(">I", read_fn(4))[0]
+        length = struct.unpack(">I", _read_except(fp, 4))[0]
     else:
         raise Exception("logic error, not map: 0x%02x" % ord(code))
 
     d = {}
     for i in range(length):
         # Unpack key
-        k = _unpackb(read_fn)
+        k = _unpack(fp)
 
         if isinstance(k, list):
             # Attempt to convert list into a hashable tuple
@@ -560,7 +628,7 @@ def _unpack_map(code, read_fn):
             raise DuplicateKeyException("encountered duplicate key: %s, %s" % (str(k), str(type(k))))
 
         # Unpack value
-        v = _unpackb(read_fn)
+        v = _unpack(fp)
 
         try:
             d[k] = v
@@ -568,21 +636,73 @@ def _unpack_map(code, read_fn):
             raise UnhashableKeyException("encountered unhashable key: %s" % str(k))
     return d
 
+def _unpack(fp):
+    code = _read_except(fp, 1)
+    return _unpack_dispatch_table[code](code, fp)
+
 ########################################
 
-def _byte_reader(s):
-    i = [0]
-    def read_fn(n):
-        if (i[0]+n > len(s)):
-            raise InsufficientDataException()
-        substring = s[i[0]:i[0]+n]
-        i[0] += n
-        return substring
-    return read_fn
+def _unpack2(fp):
+    """
+    Deserialize MessagePack bytes into a Python object.
 
-def _unpackb(read_fn):
-    code = read_fn(1)
-    return _unpack_dispatch_table[code](code, read_fn)
+    Args:
+        fp: a .read()-supporting file-like object
+
+    Returns:
+        A Python object.
+
+    Raises:
+        InsufficientDataException(UnpackException):
+            Insufficient data to unpack the encoded object.
+        InvalidStringException(UnpackException):
+            Invalid UTF-8 string encountered during unpacking.
+        ReservedCodeException(UnpackException):
+            Reserved code encountered during unpacking.
+        UnhashableKeyException(UnpackException):
+            Unhashable key encountered during map unpacking.
+            The serialized map cannot be deserialized into a Python dictionary.
+        DuplicateKeyException(UnpackException):
+            Duplicate key encountered during map unpacking.
+
+    Example:
+    >>> f = open("test.bin")
+    >>> umsgpack.unpackb(f)
+    {u'compact': True, u'schema': 0}
+    >>>
+    """
+    return _unpack(fp)
+
+def _unpack3(fp):
+    """
+    Deserialize MessagePack bytes into a Python object.
+
+    Args:
+        fp: a .read()-supporting file-like object
+
+    Returns:
+        A Python object.
+
+    Raises:
+        InsufficientDataException(UnpackException):
+            Insufficient data to unpack the encoded object.
+        InvalidStringException(UnpackException):
+            Invalid UTF-8 string encountered during unpacking.
+        ReservedCodeException(UnpackException):
+            Reserved code encountered during unpacking.
+        UnhashableKeyException(UnpackException):
+            Unhashable key encountered during map unpacking.
+            The serialized map cannot be deserialized into a Python dictionary.
+        DuplicateKeyException(UnpackException):
+            Duplicate key encountered during map unpacking.
+
+    Example:
+    >>> f = open("test.bin")
+    >>> umsgpack.unpackb(f)
+    {'compact': True, 'schema': 0}
+    >>>
+    """
+    return _unpack(fp)
 
 # For Python 2, expects a str object
 def _unpackb2(s):
@@ -590,10 +710,10 @@ def _unpackb2(s):
     Deserialize MessagePack bytes into a Python object.
 
     Args:
-        s: a 'str' containing the MessagePack serialized bytes.
+        s: a 'str' containing serialized MessagePack bytes
 
     Returns:
-        A deserialized Python object.
+        A Python object.
 
     Raises:
         TypeError:
@@ -617,8 +737,7 @@ def _unpackb2(s):
     """
     if not isinstance(s, str):
         raise TypeError("packed data is not type 'str'")
-    read_fn = _byte_reader(s)
-    return _unpackb(read_fn)
+    return _unpack(io.BytesIO(s))
 
 # For Python 3, expects a bytes object
 def _unpackb3(s):
@@ -626,10 +745,10 @@ def _unpackb3(s):
     Deserialize MessagePack bytes into a Python object.
 
     Args:
-        s: a 'bytes' containing the MessagePack serialized bytes.
+        s: a 'bytes' containing serialized MessagePack bytes
 
     Returns:
-        A deserialized Python object.
+        A Python object.
 
     Raises:
         TypeError:
@@ -653,15 +772,20 @@ def _unpackb3(s):
     """
     if not isinstance(s, bytes):
         raise TypeError("packed data is not type 'bytes'")
-    read_fn = _byte_reader(s)
-    return _unpackb(read_fn)
+    return _unpack(io.BytesIO(s))
 
 ################################################################################
+### Module Initialization
+################################################################################
 
 def __init():
+    global pack
     global packb
+    global unpack
     global unpackb
+    global dump
     global dumps
+    global load
     global loads
     global compatibility
     global _float_size
@@ -678,14 +802,22 @@ def __init():
 
     # Map packb and unpackb to the appropriate version
     if sys.version_info[0] == 3:
+        pack = _pack3
         packb = _packb3
+        dump = _pack3
         dumps = _packb3
+        unpack = _unpack3
         unpackb = _unpackb3
+        load = _unpack3
         loads = _unpackb3
     else:
+        pack = _pack2
         packb = _packb2
+        dump = _pack2
         dumps = _packb2
+        unpack = _unpack2
         unpackb = _unpackb2
+        load = _unpack2
         loads = _unpackb2
 
     # Build a dispatch table for fast lookup of unpacking function
